// Author: Professor William Killian
// Date: 2019 June 4
// Class: CSCI 362.93 - Data Structures
// Assignment: 4 - Stacks and Queues
//
// Description:
// Implementation of std::queue-like container adapter

#ifndef QUEUE_HPP_
#define QUEUE_HPP_

#include <deque>

template<typename T, typename Container = std::deque<T>>
class Queue
{

public:

  using container_type = Container;
  using value_type = typename Container::value_type;
  using size_type = typename Container::size_type;
  using reference = typename Container::reference;
  using const_reference = typename Container::const_reference;

  Queue ()
    : Queue (Container ())
  {
  }

  explicit Queue (Container const& cont)
    : c (cont)
  {
  }

  explicit Queue (Container&& cont)
    : c (std::move (cont))
  {
  }

  Queue (Queue const& other) = default;
  Queue (Queue&& other) = default;
  ~Queue () = default;

  Queue&
  operator= (Queue const& other) = default;

  Queue&
  operator= (Queue&& other) = default;

  // ////////////////////////////// Element Access /////////////////////////////

  // front()
  //
  // Returns reference to the first element in the queue. This element will be the
  // first element to be removed on a call to pop().

  reference
  front ()
  {
    // TODO
  }

  const_reference
  front () const
  {
    // TODO -- hint: same as above
  }

  // back()
  //
  // Returns reference to the last element in the queue. This is the most
  // recently pushed element.

  reference
  back ()
  {
    // TODO
  }

  const_reference
  back () const
  {
    // TODO -- hint: same as above
  }

  // ////////////////////////////// Capacity /////////////////////////////

  // empty()
  //
  // Checks if the underlying container has no elements

  bool
  empty () const
  {
    // TODO
  }

  // size()
  //
  // Returns the number of elements in the underlying container

  size_type
  size () const
  {
    // TODO
  }

  // ////////////////////////////// MODIFIERS /////////////////////////////

  // push()
  //
  // Pushes the given element value to the back of the queue

  void
  push (value_type const& value)
  {
    // TODO
  }

  void
  push (value_type&& value)
  {
    // TODO -- hint: same as above but say std::move(value)
  }

  // emplace()
  //
  // Pushes a new element to the back of the queue. The element is constructed
  // in-place, i.e. no copy or move operations are performed. The constructor
  // of the element is called with exactly the same arguments as supplied to
  // the function.

  template<typename... Args>
  decltype (auto)
  emplace (Args&&... args)
  {
    return c.emplace_back (std::forward<Args> (args)...);
  }

  // pop()
  //
  // Removes the first element from the queue.

  void
  pop ()
  {
    // TODO
  }

  // swap()
  //
  // Exchanges the contents of the container adaptor with those of other.

  void
  swap (Queue& other) noexcept (std::is_nothrow_swappable_v<Container>)
  {
    using std::swap;
    swap (c, other.c);
  }

  template<typename T1, typename C1>
  friend bool
  operator== (Queue<T1, C1> const& s1, Queue<T1, C1> const& s2);

  template<typename T1, typename C1>
  friend bool
  operator< (Queue<T1, C1> const& s1, Queue<T1, C1> const& s2);

protected:
  Container c;
};

// ///////////////////////// FREE FUNCTIONS /////////////////////////

template<typename T1, typename C1>
inline bool
operator== (Queue<T1, C1> const& s1, Queue<T1, C1> const& s2)
{
  return s1.c == s2.c;
}

template<typename T1, typename C1>
inline bool
operator< (Queue<T1, C1> const& s1, Queue<T1, C1> const& s2)
{
  return s1.c < s2.c;
}

template<typename T1, typename C1>
inline bool
operator!= (Queue<T1, C1> const& s1, Queue<T1, C1> const& s2)
{
  return s1 != s2;
}

template<typename T1, typename C1>
inline bool
operator> (Queue<T1, C1> const& s1, Queue<T1, C1> const& s2)
{
  return s2 < s1;
}

template<typename T1, typename C1>
inline bool
operator>= (Queue<T1, C1> const& s1, Queue<T1, C1> const& s2)
{
  return !(s1 < s2);
}

template<typename T1, typename C1>
inline bool
operator<= (Queue<T1, C1> const& s1, Queue<T1, C1> const& s2)
{
  return !(s2 < s1);
}

namespace std
{

template<typename T, typename Container>
std::enable_if_t<std::is_swappable_v<Queue<T, Container>>>
swap (Queue<T, Container>& lhs, Queue<T, Container>& rhs) noexcept (
  std::is_nothrow_swappable_v<Container>)
{
  lhs.swap (rhs);
}

} // namespace std

#endif
