// Author: Professor William Killian
// Date: 2019 June 4
// Class: CSCI 362.93 - Data Structures
// Assignment: 4 - Stacks and Queues
//
// Description:
// Implementation of std::stack-like container adapter

#ifndef STACK_HPP_
#define STACK_HPP_

#include <deque>

template<typename T, typename Container = std::deque<T>>
class Stack
{

public:

  using container_type = Container;
  using value_type = typename Container::value_type;
  using size_type = typename Container::size_type;
  using reference = typename Container::reference;
  using const_reference = typename Container::const_reference;

  Stack ()
    : Stack (Container ())
  {
  }

  explicit Stack (Container const& cont)
    : c (cont)
  {
  }

  explicit Stack (Container&& cont)
    : c (std::move (cont))
  {
  }

  Stack (Stack const& other) = default;
  Stack (Stack&& other) = default;
  ~Stack () = default;

  Stack&
  operator= (Stack const& other) = default;

  Stack&
  operator= (Stack&& other) = default;

  // ////////////////////////////// Element Access /////////////////////////////

  // top()
  //
  // Returns reference to the top element in the stack. This is the most
  // recently pushed element. This element will be removed on a call to
  // pop().

  reference
  top ()
  {
    // TODO
  }

  const_reference
  top () const
  {
    // TODO -- hint: same as above
  }

  // ////////////////////////////// Capacity /////////////////////////////

  // empty()
  //
  // Checks if the underlying container has no elements

  bool
  empty () const
  {
    // TODO
  }

  // size()
  //
  // Returns the number of elements in the underlying container

  size_type
  size () const
  {
    // TODO
  }

  // ////////////////////////////// MODIFIERS /////////////////////////////

  // push()
  //
  // Pushes the given element value to the top of the stack.

  void
  push (value_type const& value)
  {
    // TODO
  }

  void
  push (value_type&& value)
  {
    // TODO -- hint: same as above but say std::move(value)
  }

  // emplace()
  //
  // Pushes a new element on top of the stack. The element is constructed
  // in-place, i.e. no copy or move operations are performed. The constructor
  // of the element is called with exactly the same arguments as supplied to
  // the function.

  template<typename... Args>
  decltype (auto)
  emplace (Args&&... args)
  {
    return c.emplace_back (std::forward<Args> (args)...);
  }

  // pop()
  //
  // Removes the top element from the stack

  void
  pop ()
  {
    // TODO
  }

  // swap()
  //
  // Exchanges the contents of the container adaptor with those of other.

  void
  swap (Stack& other) noexcept (std::is_nothrow_swappable_v<Container>)
  {
    using std::swap;
    swap (c, other.c);
  }

  template<typename T1, typename C1>
  friend bool
  operator== (Stack<T1, C1> const& s1, Stack<T1, C1> const& s2);

  template<typename T1, typename C1>
  friend bool
  operator< (Stack<T1, C1> const& s1, Stack<T1, C1> const& s2);

protected:
  Container c;
};

// ///////////////////////// FREE FUNCTIONS /////////////////////////

template<typename T1, typename C1>
inline bool
operator== (Stack<T1, C1> const& s1, Stack<T1, C1> const& s2)
{
  return s1.c == s2.c;
}

template<typename T1, typename C1>
inline bool
operator< (Stack<T1, C1> const& s1, Stack<T1, C1> const& s2)
{
  return s1.c < s2.c;
}
template<typename T1, typename C1>
inline bool
operator!= (Stack<T1, C1> const& s1, Stack<T1, C1> const& s2)
{
  return s1 != s2;
}
template<typename T1, typename C1>
inline bool
operator> (Stack<T1, C1> const& s1, Stack<T1, C1> const& s2)
{
  return s2 < s1;
}
template<typename T1, typename C1>
inline bool
operator>= (Stack<T1, C1> const& s1, Stack<T1, C1> const& s2)
{
  return !(s1 < s2);
}
template<typename T1, typename C1>
inline bool
operator<= (Stack<T1, C1> const& s1, Stack<T1, C1> const& s2)
{
  return !(s2 < s1);
}

namespace std
{

template<typename T, typename Container>
std::enable_if_t<std::is_swappable_v<Stack<T, Container>>>
swap (Stack<T, Container>& lhs, Stack<T, Container>& rhs) noexcept (
  std::is_nothrow_swappable_v<Container>)
{
  lhs.swap (rhs);
}

} // namespace std

#endif
